"""
Plot a Ligand Interaction Network --- :mod:`prolif.plotting.network`
====================================================================

.. versionadded:: 0.3.2

.. autoclass:: LigNetwork
   :members:

"""
import json
import re
import warnings
from collections import defaultdict
from copy import deepcopy
from html import escape

import numpy as np
import pandas as pd
from rdkit import Chem
from rdkit.Chem import rdDepictor

from ..residue import ResidueId
from ..utils import requires

try:
    from IPython.display import HTML
except ModuleNotFoundError:
    pass
else:
    warnings.filterwarnings(
        "ignore", "Consider using IPython.display.IFrame instead"  # pragma: no cover
    )


class LigNetwork:
    """Creates a ligand interaction diagram

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe with a 4-level index (ligand, protein, interaction, atom)
        and a weight column for values
    lig_mol : rdkit.Chem.rdChem.Mol
        Ligand molecule
    match3D : bool
        If ``True``, generates 2D coordines that are constrained to fit the
        3D conformation of the ligand as best as possible. Else, generate 2D
        coordinates from scratch
    kekulize : bool
        Kekulize the ligand
    molsize : int
        Multiply the coordinates by this number to create a bigger and
        more readable depiction
    rotation : int
        Rotate the structure on the XY plane
    carbon : float
        Size of the carbon atom dots on the depiction. Use `0` to hide the
        carbon dots

    Attributes
    ----------
    COLORS : dict
        Dictionnary of colors used in the diagram. Subdivided in several
        dictionaries:

        - "interactions": mapping between interactions types and colors
        - "atoms": mapping between atom symbol and colors
        - "residues": mapping between residues types and colors

    RESIDUE_TYPES : dict
        Mapping between residue names (3 letter code) and types. The types
        are then used to define how each residue should be colored.

    Notes
    -----
    You can customize the diagram by tweaking :attr:`LigNetwork.COLORS` and
    :attr:`LigNetwork.RESIDUE_TYPES` by adding or modifying the
    dictionaries inplace.
    """

    COLORS = {
        "interactions": {
            "Hydrophobic": "#59e382",
            "HBAcceptor": "#59bee3",
            "HBDonor": "#59bee3",
            "XBAcceptor": "#59bee3",
            "XBDonor": "#59bee3",
            "Cationic": "#e35959",
            "Anionic": "#5979e3",
            "CationPi": "#e359d8",
            "PiCation": "#e359d8",
            "PiStacking": "#b559e3",
            "EdgeToFace": "#b559e3",
            "FaceToFace": "#b559e3",
            "VdWContact": "#59e3ad",
        },
        "atoms": {
            "C": "black",
            "N": "blue",
            "O": "red",
            "S": "#dece1b",
            "P": "orange",
            "F": "lime",
            "Cl": "lime",
            "Br": "lime",
            "I": "lime",
        },
        "residues": {
            "Aliphatic": "#59e382",
            "Aromatic": "#b559e3",
            "Acidic": "#e35959",
            "Basic": "#5979e3",
            "Polar": "#59bee3",
            "Sulfur": "#e3ce59",
        },
    }
    RESIDUE_TYPES = {
        "ALA": "Aliphatic",
        "GLY": "Aliphatic",
        "ILE": "Aliphatic",
        "LEU": "Aliphatic",
        "PRO": "Aliphatic",
        "VAL": "Aliphatic",
        "PHE": "Aromatic",
        "TRP": "Aromatic",
        "TYR": "Aromatic",
        "ASP": "Acidic",
        "GLU": "Acidic",
        "ARG": "Basic",
        "HIS": "Basic",
        "HID": "Basic",
        "HIE": "Basic",
        "HIP": "Basic",
        "HSD": "Basic",
        "HSE": "Basic",
        "HSP": "Basic",
        "LYS": "Basic",
        "SER": "Polar",
        "THR": "Polar",
        "ASN": "Polar",
        "GLN": "Polar",
        "CYS": "Sulfur",
        "CYM": "Sulfur",
        "CYX": "Sulfur",
        "MET": "Sulfur",
    }
    _LIG_PI_INTERACTIONS = ["EdgeToFace", "FaceToFace", "PiStacking", "PiCation"]
    _JS_TEMPLATE = """
        var ifp, legend, nodes, edges, legend_buttons;
        function drawGraph(_id, nodes, edges, options) {
            var container = document.getElementById(_id);
            nodes = new vis.DataSet(nodes);
            edges = new vis.DataSet(edges);
            var data = {nodes: nodes, edges: edges};
            var network = new vis.Network(container, data, options);
            network.on("stabilizationIterationsDone", function () {
                network.setOptions( { physics: false } );
            });
            return network;
        }
        nodes = %(nodes)s;
        edges = %(edges)s;
        ifp = drawGraph('%(div_id)s', nodes, edges, %(options)s);
    """
    _HTML_TEMPLATE = """
        <html>
        <head>
        <script type="text/javascript" src="https://unpkg.com/vis-network@9.0.4/dist/vis-network.min.js"></script>
        <link href="https://unpkg.com/vis-network@9.0.4/dist/dist/vis-network.min.css" rel="stylesheet" type="text/css" />
        <style type="text/css">
            body {
                padding: 0;
                margin: 0;
                background: #fff;
            }
            .legend-btn.residues.disabled {
                background: #b4b4b4 !important;
                color: #555 !important;
            }
            .legend-btn.interactions.disabled {
                border-color: #b4b4b4 !important;
                color: #555 !important;
            }
        </style>
        </head>
        <body>
        <div id="mynetwork"></div>
        <div id="networklegend"></div>
        <script type="text/javascript">
            %(js)s
        </script>
        </body>
        </html>
    """

    def __init__(
        self,
        df,
        lig_mol,
        match3D=True,
        kekulize=False,
        molsize=35,
        rotation=0,
        carbon=0.16,
    ):
        self.df = df
        mol = deepcopy(lig_mol)
        Chem.SanitizeMol(mol, Chem.SanitizeFlags.SANITIZE_SETAROMATICITY)
        self._ring_info = mol.GetRingInfo()
        if kekulize:
            Chem.Kekulize(mol)
        if match3D:
            rdDepictor.GenerateDepictionMatching3DStructure(mol, lig_mol)
        else:
            rdDepictor.Compute2DCoords(mol, clearConfs=True)
        xyz = mol.GetConformer().GetPositions()
        if rotation:
            theta = np.radians(rotation)
            c, s = np.cos(theta), np.sin(theta)
            R = np.array([[c, s], [-s, c]])
            xy, z = xyz[:, :2], xyz[:, 2:3]
            center = xy.mean(axis=0)
            xy = ((xy - center) @ R.T) + center
            xyz = np.concatenate([xy, z], axis=1)
        if carbon:
            self._carbon = {
                "label": " ",
                "shape": "dot",
                "color": self.COLORS["atoms"]["C"],
                "size": molsize * carbon,
            }
        else:
            self._carbon = {"label": " ", "shape": "text"}
        self.xyz = molsize * xyz
        self.mol = mol
        self._multiplier = molsize
        self.options = {}
        self._max_interaction_width = 6
        self._avoidOverlap = 0.8
        self._springConstant = 0.1
        self._bond_color = "black"
        self._default_atom_color = "grey"
        self._default_residue_color = "#dbdbdb"
        self._default_interaction_color = "#dbdbdb"
        # regroup interactions of the same color
        temp = defaultdict(list)
        interactions = set(df.index.get_level_values("interaction").unique())
        for interaction in interactions:
            color = self.COLORS["interactions"].get(
                interaction, self._default_interaction_color
            )
            temp[color].append(interaction)
        self._interaction_types = {
            interaction: "/".join(interaction_group)
            for interaction_group in temp.values()
            for interaction in interaction_group
        }

    @classmethod
    def from_ifp(cls, ifp, lig, kind="aggregate", frame=0, threshold=0.3, **kwargs):
        """Helper method to create a ligand interaction diagram from an IFP
        DataFrame obtained with ``fp.to_dataframe(return_atoms=True)``

        Notes
        -----
        Two kinds of diagrams can be rendered: either for a designated frame or
        by aggregating the results on the whole IFP and optionnally discarding
        interactions that occur less frequently than a threshold. In the latter
        case (aggregate), only the most frequent ligand atom interaction is
        rendered.

        Parameters
        ----------
        ifp : pandas.DataFrame
            The result of ``fp.to_dataframe(return_atoms=True)``
        lig : rdkit.Chem.rdChem.Mol
            Ligand molecule
        kind : str
            One of "aggregate" or "frame"
        frame : int or str
            Frame number, as read in ``ifp.index``. Only applicable for
            ``kind="frame"``
        threshold : float
            Frequency threshold, between 0 and 1. Only applicable for
            ``kind="aggregate"``
        kwargs : object
            Other arguments passed to the :class:`LigNetwork` class
        """
        if kind == "aggregate":
            data = (
                pd.get_dummies(
                    ifp.applymap(lambda x: x[0]).astype(object), prefix_sep=", "
                )
                .rename(columns=lambda x: x.translate({ord(c): None for c in "()'"}))
                .mean()
            )
            index = [i.split(", ") for i in data.index]
            index = [[j for j in i[:-1] + [int(float(i[-1]))]] for i in index]
            data.index = pd.MultiIndex.from_tuples(
                index, names=["ligand", "protein", "interaction", "atom"]
            )
            data = data.to_frame()
            data.rename(columns={data.columns[-1]: "weight"}, inplace=True)
            # merge different ligand atoms before applying the threshold
            data = data.join(
                data.groupby(level=["ligand", "protein", "interaction"]).sum(),
                rsuffix="_total",
            )
            # threshold and keep most occuring atom
            data = (
                data.loc[data["weight_total"] >= threshold]
                .drop(columns="weight_total")
                .sort_values("weight", ascending=False)
                .groupby(level=["ligand", "protein", "interaction"])
                .head(1)
                .sort_index()
            )
            return cls(data, lig, **kwargs)
        elif kind == "frame":
            data = (
                ifp.loc[ifp.index == frame]
                .T.applymap(lambda x: x[0])
                .dropna()
                .astype(int)
                .reset_index()
            )
            data.rename(columns={data.columns[-1]: "atom"}, inplace=True)
            data["weight"] = 1
            data.set_index(["ligand", "protein", "interaction", "atom"], inplace=True)
            return cls(data, lig, **kwargs)
        else:
            raise ValueError(f'{kind!r} must be "aggregate" or "frame"')

    def _make_carbon(self):
        return deepcopy(self._carbon)

    def _make_lig_node(self, atom):
        """Prepare ligand atoms"""
        idx = atom.GetIdx()
        elem = atom.GetSymbol()
        if elem == "H" and idx not in self.df.index.get_level_values("atom"):
            self.exclude.append(idx)
            return
        charge = atom.GetFormalCharge()
        if charge != 0:
            charge = "{}{}".format(
                "" if abs(charge) == 1 else str(charge), "+" if charge > 0 else "-"
            )
            label = f"{elem}{charge}"
            shape = "ellipse"
        else:
            label = elem
            shape = "circle"
        if elem == "C":
            node = self._make_carbon()
        else:
            node = {
                "label": label,
                "shape": shape,
                "color": "white",
                "font": {
                    "color": self.COLORS["atoms"].get(elem, self._default_atom_color)
                },
            }
        node.update(
            {
                "id": idx,
                "x": float(self.xyz[idx, 0]),
                "y": float(self.xyz[idx, 1]),
                "fixed": True,
                "group": "ligand",
                "borderWidth": 0,
            }
        )
        self.nodes[idx] = node

    def _make_lig_edge(self, bond):
        """Prepare ligand bonds"""
        idx = [bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()]
        if any(i in self.exclude for i in idx):
            return
        btype = bond.GetBondTypeAsDouble()
        if btype == 1:
            self.edges.append(
                {
                    "from": idx[0],
                    "to": idx[1],
                    "color": self._bond_color,
                    "physics": False,
                    "group": "ligand",
                    "width": 4,
                }
            )
        else:
            self._make_non_single_bond(idx, btype)

    def _make_non_single_bond(self, ids, btype, bdist=0.06, dash=[10]):
        """Prepare double, triple and aromatic bonds"""
        xyz = self.xyz[ids]
        d = xyz[1, :2] - xyz[0, :2]
        length = np.sqrt((d**2).sum())
        u = d / length
        p = np.array([-u[1], u[0]])
        nodes = []
        dist = bdist * self._multiplier * np.ceil(btype)
        dashes = False if btype in [2, 3] else dash
        for perp in (p, -p):
            for point in xyz:
                xy = point[:2] + perp * dist
                _id = hash(xy.tobytes())
                nodes.append(_id)
                self.nodes[_id] = {
                    "id": _id,
                    "x": xy[0],
                    "y": xy[1],
                    "shape": "text",
                    "label": " ",
                    "fixed": True,
                    "physics": False,
                }
        l1, l2, r1, r2 = nodes
        self.edges.extend(
            [
                {
                    "from": l1,
                    "to": l2,
                    "color": self._bond_color,
                    "physics": False,
                    "dashes": dashes,
                    "group": "ligand",
                    "width": 4,
                },
                {
                    "from": r1,
                    "to": r2,
                    "color": self._bond_color,
                    "physics": False,
                    "dashes": dashes,
                    "group": "ligand",
                    "width": 4,
                },
            ]
        )
        if btype == 3:
            self.edges.append(
                {
                    "from": ids[0],
                    "to": ids[1],
                    "color": self._bond_color,
                    "physics": False,
                    "group": "ligand",
                    "width": 4,
                }
            )

    def _make_interactions(self, mass=2):
        """Prepare lig-prot interactions"""
        restypes = {}
        for prot_res in self.df.index.get_level_values("protein").unique():
            resname = ResidueId.from_string(prot_res).name
            restype = self.RESIDUE_TYPES.get(resname)
            restypes[prot_res] = restype
            color = self.COLORS["residues"].get(restype, self._default_residue_color)
            node = {
                "id": prot_res,
                "label": prot_res,
                "color": color,
                "shape": "box",
                "borderWidth": 0,
                "physics": True,
                "mass": mass,
                "group": "protein",
                "residue_type": restype,
            }
            self.nodes[prot_res] = node
        for ((lig_res, prot_res, interaction, lig_id), (weight,)) in self.df.iterrows():
            if interaction in self._LIG_PI_INTERACTIONS:
                centroid = self._get_ring_centroid(lig_id)
                origin = str((lig_res, prot_res, interaction))
                self.nodes[origin] = {
                    "id": origin,
                    "x": centroid[0],
                    "y": centroid[1],
                    "shape": "text",
                    "label": " ",
                    "fixed": True,
                    "physics": False,
                }
            else:
                origin = int(lig_id)
            edge = {
                "from": origin,
                "to": prot_res,
                "title": interaction,
                "interaction_type": self._interaction_types.get(
                    interaction, interaction
                ),
                "color": self.COLORS["interactions"].get(
                    interaction, self._default_interaction_color
                ),
                "smooth": {"type": "cubicBezier", "roundness": 0.2},
                "dashes": [10],
                "width": weight * self._max_interaction_width,
                "group": "interaction",
            }
            self.edges.append(edge)

    def _get_ring_centroid(self, index):
        """Find ring coordinates using the index of one of the ring atoms"""
        for r in self._ring_info.AtomRings():
            if index in r:
                break
        else:
            raise ValueError(
                "No ring containing this atom index was found in " "the given molecule"
            )
        return self.xyz[list(r)].mean(axis=0)

    def _patch_hydrogens(self):
        """Patch hydrogens on heteroatoms

        Hydrogen atoms that aren't part of any interaction have been hidden at
        this stage, but they should be added to the label of the heteroatom for
        clarity
        """
        to_patch = defaultdict(int)
        for idx in self.exclude:
            h = self.mol.GetAtomWithIdx(idx)
            atom = h.GetNeighbors()[0]
            if atom.GetSymbol() != "C":
                to_patch[atom.GetIdx()] += 1
        for idx, nH in to_patch.items():
            node = self.nodes[idx]
            h_str = "H" if nH == 1 else f"H{nH}"
            label = re.sub(r"(\w+)(.*)", rf"\1{h_str}\2", node["label"])
            node["label"] = label
            node["shape"] = "ellipse"

    def _make_graph_data(self):
        """Prepares the nodes and edges"""
        self.exclude = []
        self.nodes = {}
        self.edges = []
        # show residues
        self._make_interactions()
        # show ligand
        for atom in self.mol.GetAtoms():
            self._make_lig_node(atom)
        for bond in self.mol.GetBonds():
            self._make_lig_edge(bond)
        self._patch_hydrogens()
        self.nodes = list(self.nodes.values())

    def _get_js(self, width="100%", height="500px", div_id="mynetwork", fontsize=20):
        """Returns the JavaScript code to draw the network"""
        self.width = width
        self.height = height
        self._make_graph_data()
        options = {
            "width": width,
            "height": height,
            "nodes": {
                "font": {"size": fontsize},
            },
            "physics": {
                "barnesHut": {
                    "avoidOverlap": self._avoidOverlap,
                    "springConstant": self._springConstant,
                }
            },
        }
        options.update(self.options)
        js = self._JS_TEMPLATE % dict(
            div_id=div_id,
            nodes=json.dumps(self.nodes),
            edges=json.dumps(self.edges),
            options=json.dumps(options),
        )
        js += self._get_legend()
        return js

    def _get_html(self, **kwargs):
        """Returns the HTML code to draw the network"""
        return self._HTML_TEMPLATE % dict(js=self._get_js(**kwargs))

    def _get_legend(self, height="90px"):
        available = {}
        buttons = []
        map_color_restype = {c: t for t, c in self.COLORS["residues"].items()}
        map_color_interactions = {
            self.COLORS["interactions"].get(i, self._default_interaction_color): t
            for i, t in self._interaction_types.items()
        }
        # residues
        for node in self.nodes:
            if node.get("group", "") == "protein":
                color = node["color"]
                available[color] = map_color_restype.get(color, "Unknown")
        available = {
            k: v for k, v in sorted(available.items(), key=lambda item: item[1])
        }
        for i, (color, restype) in enumerate(available.items()):
            buttons.append(
                {"index": i, "label": restype, "color": color, "group": "residues"}
            )
        # interactions
        available.clear()
        for edge in self.edges:
            if edge.get("group", "") == "interaction":
                color = edge["color"]
                available[color] = map_color_interactions[color]
        available = {
            k: v for k, v in sorted(available.items(), key=lambda item: item[1])
        }
        for i, (color, interaction) in enumerate(available.items()):
            buttons.append(
                {
                    "index": i,
                    "label": interaction,
                    "color": color,
                    "group": "interactions",
                }
            )
        # JS code
        if all("px" in h for h in [self.height, height]):
            h1 = int(re.findall(r"(\d+)\w+", self.height)[0])
            h2 = int(re.findall(r"(\d+)\w+", height)[0])
            self.height = f"{h1+h2}px"
        return """
        legend_buttons = %(buttons)s;
        legend = document.getElementById('%(div_id)s');
        var div_residues = document.createElement('div');
        var div_interactions = document.createElement('div');
        var disabled = [];
        var legend_callback = function() {
            this.classList.toggle("disabled");
            var hide = this.classList.contains("disabled");
            var show = !hide;
            var btn_label = this.innerHTML;
            if (hide) {
                disabled.push(btn_label);
            } else {
                disabled = disabled.filter(x => x !== btn_label);
            }
            var node_update = [],
                edge_update = [];
            // click on residue type
            if (this.classList.contains("residues")) {
                nodes.forEach((node) => {
                    // find nodes corresponding to this type
                    if (node.residue_type === btn_label) {
                        // if hiding this type and residue isn't already hidden
                        if (hide && !node.hidden) {
                            node.hidden = true;
                            node_update.push(node);
                        // if showing this type and residue isn't already visible
                        } else if (show && node.hidden) {
                            // display if there's at least one of its edge that isn't hidden
                            num_edges_active = edges.filter(x => x.to === node.id)
                                                    .map(x => Boolean(x.hidden))
                                                    .filter(x => !x)
                                                    .length;
                            if (num_edges_active > 0) {
                                node.hidden = false;
                                node_update.push(node);
                            }
                        }
                    }
                });
                ifp.body.data.nodes.update(node_update);
            // click on interaction type
            } else {
                edges.forEach((edge) => {
                    // find edges corresponding to this type
                    if (edge.interaction_type === btn_label) {
                        edge.hidden = !edge.hidden;
                        edge_update.push(edge);
                        // number of active edges for the corresponding residue
                        var num_edges_active = edges.filter(x => x.to === edge.to)
                                               .map(x => Boolean(x.hidden))
                                               .filter(x => !x)
                                               .length;
                        // find corresponding residue
                        var ix = nodes.findIndex(x => x.id === edge.to);
                        // only change visibility if residue_type not being hidden
                        if (!(disabled.includes(nodes[ix].residue_type))) {
                            // hide if no edge being shown for this residue
                            if (hide && (num_edges_active === 0)) {
                                nodes[ix].hidden = true;
                                node_update.push(nodes[ix]);
                            // show if edges are being shown
                            } else if (show && (num_edges_active > 0)) {
                                nodes[ix].hidden = false;
                                node_update.push(nodes[ix]);
                            }
                        }
                    }
                });
                ifp.body.data.nodes.update(node_update);
                ifp.body.data.edges.update(edge_update);
            }
        };
        legend_buttons.forEach(function(v,i) {
            if (v.group === "residues") {
                var div = div_residues;
                var border = "none";
                var color = v.color;
            } else {
                var div = div_interactions;
                var border = "3px dashed " + v.color;
                var color = "white";
            }
            var button = div.appendChild(document.createElement('button'));
            button.classList.add("legend-btn", v.group);
            button.innerHTML = v.label;
            Object.assign(button.style, {
                "cursor": "pointer",
                "background-color": color,
                "border": border,
                "border-radius": "5px",
                "padding": "5px",
                "margin": "5px",
                "font": "14px 'Arial', sans-serif",
            });
            button.onclick = legend_callback;
        });
        legend.appendChild(div_residues);
        legend.appendChild(div_interactions);
        """ % dict(
            div_id="networklegend", buttons=json.dumps(buttons)
        )

    @requires("IPython.display")
    def display(self, **kwargs):
        """Prepare and display the network"""
        html = self._get_html(**kwargs)
        iframe = (
            '<iframe width="{width}" height="{height}" frameborder="0" '
            'srcdoc="{doc}"></iframe>'
        )
        return HTML(
            iframe.format(width=self.width, height=self.height, doc=escape(html))
        )

    @requires("IPython.display")
    def show(self, filename, **kwargs):
        """Save the network as HTML and display the resulting file"""
        html = self._get_html(**kwargs)
        with open(filename, "w") as f:
            f.write(html)
        iframe = (
            '<iframe width="{width}" height="{height}" frameborder="0" '
            'src="{filename}"></iframe>'
        )
        return HTML(
            iframe.format(width=self.width, height=self.height, filename=filename)
        )

    def save(self, fp, **kwargs):
        """Save the network to an HTML file

        Parameters
        ----------
        fp : str or file-like object
            Name of the output file, or file-like object
        """
        html = self._get_html(**kwargs)
        try:
            fp.write(html)
        except AttributeError:
            with open(fp, "w") as f:
                f.write(html)



import os
from contextlib import contextmanager
from io import StringIO
from tempfile import NamedTemporaryFile

import MDAnalysis as mda
import pytest

import prolif as plf
from prolif.plotting.network import LigNetwork


@contextmanager
def TempFilename(mode="w"):
    f = NamedTemporaryFile(delete=False, mode=mode)
    try:
        f.close()
        yield f.name
    finally:
        os.unlink(f.name)


class TestLigNetwork:
    @pytest.fixture(scope="class")
    def lignetwork_data(self):
        u = mda.Universe(plf.datafiles.TOP)
        prot = u.select_atoms("protein and resid 200:331")
        lig = u.select_atoms("resname LIG")
        fp = plf.Fingerprint()
        fp.run(u.trajectory, lig, prot)
        df = fp.to_dataframe(return_atoms=True)
        lig = plf.Molecule.from_mda(lig)
        return lig, df

    def test_integration_frame(self, lignetwork_data):
        lig, df = lignetwork_data
        net = LigNetwork.from_ifp(df, lig, kind="frame", frame=0)
        with StringIO() as buffer:
            net.save(buffer)
            buffer.seek(0)
            html = buffer.read()
        assert "PHE331.B" in html

    def test_integration_agg(self, lignetwork_data):
        lig, df = lignetwork_data
        net = LigNetwork.from_ifp(df, lig, kind="aggregate", threshold=0)
        with StringIO() as buffer:
            net.save(buffer)
            buffer.seek(0)
            html = buffer.read()
        assert "PHE331.B" in html

    def test_kwargs(self, lignetwork_data):
        lig, df = lignetwork_data
        net = LigNetwork.from_ifp(
            df, lig, kekulize=True, match3D=False, rotation=42, carbon=0
        )
        with StringIO() as buffer:
            net.save(buffer)
            buffer.seek(0)
            html = buffer.read()
        assert "PHE331.B" in html

    def test_save_file(self, lignetwork_data):
        lig, df = lignetwork_data
        net = LigNetwork.from_ifp(df, lig)
        with TempFilename("w+") as filename:
            net.save(filename)
            with open(filename, "r") as f:
                assert "PHE331.B" in f.read()

    def test_from_ifp_raises_kind(self, lignetwork_data):
        lig, df = lignetwork_data
        with pytest.raises(ValueError, match='must be "aggregate" or "frame"'):
            LigNetwork.from_ifp(df, lig, kind="foo")



"""
Detecting interactions between residues --- :mod:`prolif.interactions`
======================================================================

You can declare your own interaction class like this::

    from scipy.spatial import distance_matrix

    class CloseContact(prolif.Interaction):
        def detect(self, res1, res2, threshold=2.0):
            dist_matrix = distance_matrix(res1.xyz, res2.xyz)
            if (dist_matrix <= threshold).any():
                # return format: bool, ligand atom index, protein atom index
                return True, None, None
            return False, None, None

.. warning:: Your custom class must inherit from :class:`prolif.interactions.Interaction`

The new "CloseContact" class is then automatically added to the list of
interactions available to the fingerprint generator::

    >>> u = mda.Universe(prolif.datafiles.TOP, prolif.datafiles.TRAJ)
    >>> prot = u.select_atoms("protein")
    >>> lig = u.select_atoms("resname LIG")
    >>> fp = prolif.Fingerprint(interactions="all")
    >>> df = fp.run(u.trajectory[0:1], lig, prot).to_dataframe()
    >>> df.xs("CloseContact", level=1, axis=1)
       ASP129.A  VAL201.A
    0         1         1
    >>> lmol = prolif.Molecule.from_mda(lig)
    >>> pmol = prolif.Molecule.from_mda(prot)
    >>> fp.closecontact(lmol, pmol["ASP129.A"])
    True

Note that some of the SMARTS patterns used in the interaction classes are inspired from
`Pharmit`_ and `RDKit`_.

.. _Pharmit: https://sourceforge.net/p/pharmit/code/ci/master/tree/src/pharmarec.cpp
.. _RDKit: https://github.com/rdkit/rdkit/blob/master/Data/BaseFeatures.fdef

"""

import warnings
from abc import ABC, ABCMeta, abstractmethod
from itertools import product
from math import radians

import numpy as np
from MDAnalysis.topology.tables import vdwradii
from rdkit import Geometry
from rdkit.Chem import MolFromSmarts

from .utils import angle_between_limits, get_centroid, get_ring_normal_vector

_INTERACTIONS = {}


class _InteractionMeta(ABCMeta):
    """Metaclass to register interactions automatically"""

    def __init__(cls, name, bases, classdict):
        type.__init__(cls, name, bases, classdict)
        if name in _INTERACTIONS.keys():
            warnings.warn(
                f"The {name!r} interaction has been superseded by a "
                f"new class with id {id(cls):#x}"
            )
        _INTERACTIONS[name] = cls


class Interaction(ABC, metaclass=_InteractionMeta):
    """Abstract class for interactions

    All interaction classes must inherit this class and define a
    :meth:`~detect` method
    """

    @abstractmethod
    def detect(self, **kwargs):
        pass


def get_mapindex(res, index):
    """Get the index of the atom in the original molecule

    Parameters
    ----------
    res : prolif.residue.Residue
        The residue in the protein or ligand
    index : int
        The index of the atom in the :class:`~prolif.residue.Residue`

    Returns
    -------
    mapindex : int
        The index of the atom in the :class:`~prolif.molecule.Molecule`
    """
    return res.GetAtomWithIdx(index).GetUnsignedProp("mapindex")


class _Distance(Interaction):
    """Generic class for distance-based interactions

    Parameters
    ----------
    lig_pattern : str
        SMARTS pattern for atoms in ligand residues
    prot_pattern : str
        SMARTS pattern for atoms in protein residues
    distance : float
        Cutoff distance, measured between the first atom of each pattern
    """

    def __init__(self, lig_pattern, prot_pattern, distance):
        self.lig_pattern = MolFromSmarts(lig_pattern)
        self.prot_pattern = MolFromSmarts(prot_pattern)
        self.distance = distance

    def detect(self, lig_res, prot_res):
        lig_matches = lig_res.GetSubstructMatches(self.lig_pattern)
        prot_matches = prot_res.GetSubstructMatches(self.prot_pattern)
        if lig_matches and prot_matches:
            for lig_match, prot_match in product(lig_matches, prot_matches):
                alig = Geometry.Point3D(*lig_res.xyz[lig_match[0]])
                aprot = Geometry.Point3D(*prot_res.xyz[prot_match[0]])
                if alig.Distance(aprot) <= self.distance:
                    return True, lig_match[0], prot_match[0]
        return False, None, None


class Hydrophobic(_Distance):
    """Hydrophobic interaction

    Parameters
    ----------
    hydrophobic : str
        SMARTS query for hydrophobic atoms
    distance : float
        Cutoff distance for the interaction


    .. versionchanged:: 1.1.0
        The initial SMARTS pattern was too broad.

    """

    def __init__(
        self,
        hydrophobic=(
            "[c,s,Br,I,S&H0&v2," "$([D3,D4;#6])&!$([#6]~[#7,#8,#9])&!$([#6X4H0]);+0]"
        ),
        distance=4.5,
    ):
        super().__init__(hydrophobic, hydrophobic, distance)


class _BaseHBond(Interaction):
    """Base class for Hydrogen bond interactions

    Parameters
    ----------
    donor : str
        SMARTS for ``[Donor]-[Hydrogen]``
    acceptor : str
        SMARTS for ``[Acceptor]``
    distance : float
        Cutoff distance between the donor and acceptor atoms
    angles : tuple
        Min and max values for the ``[Donor]-[Hydrogen]...[Acceptor]`` angle


    .. versionchanged:: 1.1.0
        The initial SMARTS pattern was too broad.

    """

    def __init__(
        self,
        donor="[$([O,S;+0]),$([N;v3,v4&+1]),n+0]-[H]",
        acceptor=(
            "[#7&!$([nX3])&!$([NX3]-*=[O,N,P,S])&!$([NX3]-[a])&!$([Nv4&+1]),"
            "O&!$([OX2](C)C=O)&!$(O(~a)~a)&!$(O=N-*)&!$([O-]-N=O),o+0,"
            "F&$(F-[#6])&!$(F-[#6][F,Cl,Br,I])]"
        ),
        distance=3.5,
        angles=(130, 180),
    ):
        self.donor = MolFromSmarts(donor)
        self.acceptor = MolFromSmarts(acceptor)
        self.distance = distance
        self.angles = tuple(radians(i) for i in angles)

    def detect(self, acceptor, donor):
        acceptor_matches = acceptor.GetSubstructMatches(self.acceptor)
        donor_matches = donor.GetSubstructMatches(self.donor)
        if acceptor_matches and donor_matches:
            for donor_match, acceptor_match in product(donor_matches, acceptor_matches):
                # D-H ... A
                d = Geometry.Point3D(*donor.xyz[donor_match[0]])
                h = Geometry.Point3D(*donor.xyz[donor_match[1]])
                a = Geometry.Point3D(*acceptor.xyz[acceptor_match[0]])
                if d.Distance(a) <= self.distance:
                    hd = h.DirectionVector(d)
                    ha = h.DirectionVector(a)
                    # get DHA angle
                    angle = hd.AngleTo(ha)
                    if angle_between_limits(angle, *self.angles):
                        return True, acceptor_match[0], donor_match[1]
        return False, None, None


class HBDonor(_BaseHBond):
    """Hbond interaction between a ligand (donor) and a residue (acceptor)"""

    def detect(self, ligand, residue):
        bit, ires, ilig = super().detect(residue, ligand)
        return bit, ilig, ires


class HBAcceptor(_BaseHBond):
    """Hbond interaction between a ligand (acceptor) and a residue (donor)"""

    def detect(self, ligand, residue):
        return super().detect(ligand, residue)


class _BaseXBond(Interaction):
    """Base class for Halogen bond interactions

    Parameters
    ----------
    donor : str
        SMARTS for ``[Donor]-[Halogen]``
    acceptor : str
        SMARTS for ``[Acceptor]-[R]``
    distance : float
        Cutoff distance between the halogen and acceptor atoms
    axd_angles : tuple
        Min and max values for the ``[Acceptor]...[Halogen]-[Donor]`` angle
    xar_angles : tuple
        Min and max values for the ``[R]-[Acceptor]...[Halogen]`` angle

    Notes
    -----
    Distance and angle adapted from Auffinger et al. PNAS 2004
    """

    def __init__(
        self,
        donor="[#6,#7,Si,F,Cl,Br,I]-[Cl,Br,I,At]",
        acceptor="[#7,#8,P,S,Se,Te,a;!+{1-}][*]",
        distance=3.5,
        axd_angles=(130, 180),
        xar_angles=(80, 140),
    ):
        self.donor = MolFromSmarts(donor)
        self.acceptor = MolFromSmarts(acceptor)
        self.distance = distance
        self.axd_angles = tuple(radians(i) for i in axd_angles)
        self.xar_angles = tuple(radians(i) for i in xar_angles)

    def detect(self, acceptor, donor):
        acceptor_matches = acceptor.GetSubstructMatches(self.acceptor)
        donor_matches = donor.GetSubstructMatches(self.donor)
        if acceptor_matches and donor_matches:
            for donor_match, acceptor_match in product(donor_matches, acceptor_matches):
                # D-X ... A distance
                d = Geometry.Point3D(*donor.xyz[donor_match[0]])
                x = Geometry.Point3D(*donor.xyz[donor_match[1]])
                a = Geometry.Point3D(*acceptor.xyz[acceptor_match[0]])
                if x.Distance(a) <= self.distance:
                    # D-X ... A angle
                    xd = x.DirectionVector(d)
                    xa = x.DirectionVector(a)
                    angle = xd.AngleTo(xa)
                    if angle_between_limits(angle, *self.axd_angles):
                        # X ... A-R angle
                        r = Geometry.Point3D(*acceptor.xyz[acceptor_match[1]])
                        ax = a.DirectionVector(x)
                        ar = a.DirectionVector(r)
                        angle = ax.AngleTo(ar)
                        if angle_between_limits(angle, *self.xar_angles):
                            return True, acceptor_match[0], donor_match[1]
        return False, None, None


class XBAcceptor(_BaseXBond):
    """Halogen bonding between a ligand (acceptor) and a residue (donor)"""

    def detect(self, ligand, residue):
        return super().detect(ligand, residue)


class XBDonor(_BaseXBond):
    """Halogen bonding between a ligand (donor) and a residue (acceptor)"""

    def detect(self, ligand, residue):
        bit, ires, ilig = super().detect(residue, ligand)
        return bit, ilig, ires


class _BaseIonic(_Distance):
    """Base class for ionic interactions

    .. versionchanged:: 1.1.0
        Handles resonance forms for common acids, amidine and guanidine.

    """

    def __init__(
        self,
        cation="[+{1-},$([NX3&!$([NX3]-O)]-[C]=[NX3+])]",
        anion="[-{1-},$(O=[C,S,P]-[O-])]",
        distance=4.5,
    ):
        super().__init__(cation, anion, distance)


class Cationic(_BaseIonic):
    """Ionic interaction between a ligand (cation) and a residue (anion)"""

    def detect(self, ligand, residue):
        return super().detect(ligand, residue)


class Anionic(_BaseIonic):
    """Ionic interaction between a ligand (anion) and a residue (cation)"""

    def detect(self, ligand, residue):
        bit, ires, ilig = super().detect(residue, ligand)
        return bit, ilig, ires


class _BaseCationPi(Interaction):
    """Base class for cation-pi interactions

    Parameters
    ----------
    cation : str
        SMARTS for cation
    pi_ring : tuple
        SMARTS for aromatic rings (5 and 6 membered rings only)
    distance : float
        Cutoff distance between the centroid and the cation
    angles : tuple
        Min and max values for the angle between the vector normal to the ring
        plane and the vector going from the centroid to the cation


    .. versionchanged:: 1.1.0
        Handles resonance forms for amidine and guanidine as cations.

    """

    def __init__(
        self,
        cation="[+{1-},$([NX3&!$([NX3]-O)]-[C]=[NX3+])]",
        pi_ring=(
            "[a;r6]1:[a;r6]:[a;r6]:[a;r6]:[a;r6]:[a;r6]:1",
            "[a;r5]1:[a;r5]:[a;r5]:[a;r5]:[a;r5]:1",
        ),
        distance=4.5,
        angles=(0, 30),
    ):
        self.cation = MolFromSmarts(cation)
        self.pi_ring = [MolFromSmarts(s) for s in pi_ring]
        self.distance = distance
        self.angles = tuple(radians(i) for i in angles)

    def detect(self, cation, pi):
        cation_matches = cation.GetSubstructMatches(self.cation)
        for pi_ring in self.pi_ring:
            pi_matches = pi.GetSubstructMatches(pi_ring)
            if not (cation_matches and pi_matches):
                continue
            for cation_match, pi_match in product(cation_matches, pi_matches):
                cat = Geometry.Point3D(*cation.xyz[cation_match[0]])
                # get coordinates of atoms matching pi-system
                pi_coords = pi.xyz[list(pi_match)]
                # centroid of pi-system as 3d point
                centroid = Geometry.Point3D(*get_centroid(pi_coords))
                # distance between cation and centroid
                if cat.Distance(centroid) > self.distance:
                    continue
                # vector normal to ring plane
                normal = get_ring_normal_vector(centroid, pi_coords)
                # vector between the centroid and the charge
                centroid_cation = centroid.DirectionVector(cat)
                # compute angle between normal to ring plane and
                # centroid-cation
                angle = normal.AngleTo(centroid_cation)
                if angle_between_limits(angle, *self.angles, ring=True):
                    return True, cation_match[0], pi_match[0]
        return False, None, None


class PiCation(_BaseCationPi):
    """Cation-Pi interaction between a ligand (aromatic ring) and a residue
    (cation)"""

    def detect(self, ligand, residue):
        bit, ires, ilig = super().detect(residue, ligand)
        return bit, ilig, ires


class CationPi(_BaseCationPi):
    """Cation-Pi interaction between a ligand (cation) and a residue
    (aromatic ring)"""

    def detect(self, ligand, residue):
        return super().detect(ligand, residue)


class _BasePiStacking(Interaction):
    """Base class for Pi-Stacking interactions

    Parameters
    ----------
    centroid_distance : float
        Cutoff distance between each rings centroid
    plane_angles : tuple
        Min and max values for the angle between the ring planes
    normal_centroids_angles : tuple
        Min and max angles allowed between the vector normal to a ring's plane,
        and the vector between the centroid of both rings.
    pi_ring : list
        List of SMARTS for aromatic rings


    .. versionchanged:: 1.1.0
        The implementation now relies on the angle between the vector normal to a ring's
        plane and the vector between centroids (``normal_centroids_angles``) instead of
        the ``shortest_distance`` parameter.

    """

    def __init__(
        self,
        centroid_distance=5.5,
        plane_angle=(0, 35),
        normal_to_centroid_angle=(0, 30),
        pi_ring=(
            "[a;r6]1:[a;r6]:[a;r6]:[a;r6]:[a;r6]:[a;r6]:1",
            "[a;r5]1:[a;r5]:[a;r5]:[a;r5]:[a;r5]:1",
        ),
    ):
        self.pi_ring = [MolFromSmarts(s) for s in pi_ring]
        self.centroid_distance = centroid_distance
        self.plane_angle = tuple(radians(i) for i in plane_angle)
        self.normal_to_centroid_angle = tuple(
            radians(i) for i in normal_to_centroid_angle
        )
        self.edge = False
        self.ring_radius = 1.7

    def detect(self, ligand, residue):
        for pi_rings in product(self.pi_ring, repeat=2):
            res_matches = residue.GetSubstructMatches(pi_rings[0])
            lig_matches = ligand.GetSubstructMatches(pi_rings[1])
            if not (lig_matches and res_matches):
                continue
            for lig_match, res_match in product(lig_matches, res_matches):
                lig_pi_coords = ligand.xyz[list(lig_match)]
                lig_centroid = Geometry.Point3D(*get_centroid(lig_pi_coords))
                res_pi_coords = residue.xyz[list(res_match)]
                res_centroid = Geometry.Point3D(*get_centroid(res_pi_coords))
                cdist = lig_centroid.Distance(res_centroid)
                if cdist > self.centroid_distance:
                    continue
                # ligand
                lig_normal = get_ring_normal_vector(lig_centroid, lig_pi_coords)
                # residue
                res_normal = get_ring_normal_vector(res_centroid, res_pi_coords)
                plane_angle = lig_normal.AngleTo(res_normal)
                if not angle_between_limits(plane_angle, *self.plane_angle, ring=True):
                    continue
                c1c2 = lig_centroid.DirectionVector(res_centroid)
                c2c1 = res_centroid.DirectionVector(lig_centroid)
                n1c1c2 = lig_normal.AngleTo(c1c2)
                n2c2c1 = res_normal.AngleTo(c2c1)
                if not (
                    angle_between_limits(
                        n1c1c2, *self.normal_to_centroid_angle, ring=True
                    )
                    or angle_between_limits(
                        n2c2c1, *self.normal_to_centroid_angle, ring=True
                    )
                ):
                    continue
                if self.edge:
                    # look for point of intersection between both ring planes
                    intersect = self._get_intersect_point(
                        lig_normal, lig_centroid, res_normal, res_centroid
                    )
                    if intersect is None:
                        continue
                    # check if intersection point falls ~within plane ring
                    intersect_dist = min(
                        lig_centroid.Distance(intersect),
                        res_centroid.Distance(intersect),
                    )
                    if intersect_dist > self.ring_radius:
                        continue
                return True, lig_match[0], res_match[0]
        return False, None, None

    @staticmethod
    def _get_intersect_point(
        plane_normal,
        plane_centroid,
        tilted_normal,
        tilted_centroid,
    ):
        # intersect line is orthogonal to both planes normal vectors
        intersect_direction = plane_normal.CrossProduct(tilted_normal)
        # setup system of linear equations to solve
        A = np.array(
            [list(plane_normal), list(tilted_normal), list(intersect_direction)]
        )
        if np.linalg.det(A) == 0:
            return None
        tilted_offset = tilted_normal.DotProduct(Geometry.Point3D(*tilted_centroid))
        plane_offset = plane_normal.DotProduct(Geometry.Point3D(*plane_centroid))
        d = np.array([[plane_offset], [tilted_offset], [0.0]])
        # point on intersect line
        point = np.linalg.solve(A, d).T[0]
        point = Geometry.Point3D(*point)
        # find projection of centroid on intersect line using vector projection
        vec = plane_centroid - point
        intersect_direction.Normalize()
        scalar_proj = intersect_direction.DotProduct(vec)
        return point + intersect_direction * scalar_proj


class FaceToFace(_BasePiStacking):
    """Face-to-face Pi-Stacking interaction between a ligand and a residue"""

    def __init__(
        self,
        centroid_distance=5.5,
        plane_angle=(0, 35),
        normal_to_centroid_angle=(0, 33),
        pi_ring=(
            "[a;r6]1:[a;r6]:[a;r6]:[a;r6]:[a;r6]:[a;r6]:1",
            "[a;r5]1:[a;r5]:[a;r5]:[a;r5]:[a;r5]:1",
        ),
    ):
        super().__init__(
            centroid_distance=centroid_distance,
            plane_angle=plane_angle,
            normal_to_centroid_angle=normal_to_centroid_angle,
            pi_ring=pi_ring,
        )


class EdgeToFace(_BasePiStacking):
    """Edge-to-face Pi-Stacking interaction between a ligand and a residue

    .. versionchanged:: 1.1.0
        In addition to the changes made to the base pi-stacking interaction, this
        implementation makes sure that the intersection between the perpendicular ring's
        plane and the other's plane falls inside the ring.

    """

    def __init__(
        self,
        centroid_distance=6.5,
        plane_angle=(50, 90),
        normal_to_centroid_angle=(0, 30),
        ring_radius=1.5,
        pi_ring=(
            "[a;r6]1:[a;r6]:[a;r6]:[a;r6]:[a;r6]:[a;r6]:1",
            "[a;r5]1:[a;r5]:[a;r5]:[a;r5]:[a;r5]:1",
        ),
    ):
        super().__init__(
            centroid_distance=centroid_distance,
            plane_angle=plane_angle,
            normal_to_centroid_angle=normal_to_centroid_angle,
            pi_ring=pi_ring,
        )
        self.edge = True
        self.ring_radius = ring_radius

    def detect(self, ligand, residue):
        return super().detect(ligand, residue)


class PiStacking(Interaction):
    """Pi-Stacking interaction between a ligand and a residue

    Parameters
    ----------
    ftf_kwargs : dict
        Parameters to pass to the underlying FaceToFace class
    etf_kwargs : dict
        Parameters to pass to the underlying EdgeToFace class


    .. versionchanged:: 0.3.4
        `shortest_distance` has been replaced by `angle_normal_centroid`

    .. versionchanged:: 1.1.0
        The implementation now directly calls :class:`EdgeToFace` and :class:`FaceToFace`
        instead of overwriting the default parameters with more generic ones.

    """

    def __init__(self, ftf_kwargs=None, etf_kwargs=None):
        self.ftf = FaceToFace(**ftf_kwargs or {})
        self.etf = EdgeToFace(**etf_kwargs or {})

    def detect(self, ligand, residue):
        ftf = self.ftf.detect(ligand, residue)
        if ftf[0] is True:
            return ftf
        return self.etf.detect(ligand, residue)


class _BaseMetallic(_Distance):
    """Base class for metal complexation

    Parameters
    ----------
    metal : str
        SMARTS for a transition metal
    ligand : str
        SMARTS for a ligand
    distance : float
        Cutoff distance


    .. versionchanged:: 1.1.0
        The initial SMARTS pattern was too broad.

    """

    def __init__(
        self,
        metal="[Ca,Cd,Co,Cu,Fe,Mg,Mn,Ni,Zn]",
        ligand="[O,#7&!$([nX3])&!$([NX3]-*=[!#6])&!$([NX3]-[a])&!$([NX4]),-{1-};!+{1-}]",
        distance=2.8,
    ):
        super().__init__(metal, ligand, distance)


class MetalDonor(_BaseMetallic):
    """Metallic interaction between a metal and a residue (chelated)"""

    def detect(self, ligand, residue):
        return super().detect(ligand, residue)


class MetalAcceptor(_BaseMetallic):
    """Metallic interaction between a ligand (chelated) and a metal residue"""

    def detect(self, ligand, residue):
        bit, ires, ilig = super().detect(residue, ligand)
        return bit, ilig, ires


class VdWContact(Interaction):
    """Interaction based on the van der Waals radii of interacting atoms.

    Parameters
    ----------
    tolerance : float
        Tolerance added to the sum of vdW radii of atoms before comparing to
        the interatomic distance. If ``distance <= sum_vdw + tolerance`` the
        atoms are identified as a contact

    Raises
    ------
    ValueError : ``tolerance`` parameter cannot be negative
    """

    def __init__(self, tolerance=0.0):
        if tolerance >= 0:
            self.tolerance = tolerance
        else:
            raise ValueError("`tolerance` must be 0 or positive")
        self._vdw_cache = {}

    def detect(self, ligand, residue):
        lxyz = ligand.GetConformer()
        rxyz = residue.GetConformer()
        for la, ra in product(ligand.GetAtoms(), residue.GetAtoms()):
            lig = la.GetSymbol().upper()
            res = ra.GetSymbol().upper()
            try:
                vdw = self._vdw_cache[frozenset((lig, res))]
            except KeyError:
                vdw = vdwradii[lig] + vdwradii[res] + self.tolerance
                self._vdw_cache[frozenset((lig, res))] = vdw
            dist = lxyz.GetAtomPosition(la.GetIdx()).Distance(
                rxyz.GetAtomPosition(ra.GetIdx())
            )
            if dist <= vdw:
                return True, la.GetIdx(), ra.GetIdx()
        return False, None, None



"""
Helper functions --- :mod:`prolif.utils`
========================================
"""
import warnings
from collections import defaultdict
from collections.abc import Iterable
from contextlib import contextmanager
from copy import deepcopy
from functools import wraps
from importlib.util import find_spec
from math import pi

import numpy as np
import pandas as pd
from rdkit import rdBase
from rdkit.Chem import FragmentOnBonds, GetMolFrags, SplitMolByPDBResidues
from rdkit.DataStructs import ExplicitBitVect
from rdkit.Geometry import Point3D
from scipy.spatial import cKDTree

from .residue import ResidueId

_90_deg_to_rad = pi / 2


def requires(module):  # pragma: no cover
    def inner(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if find_spec(module):
                return func(*args, **kwargs)
            raise ModuleNotFoundError(
                f"The module {module!r} is required to use {func.__name__!r} "
                "but it is not installed!"
            )

        return wrapper

    return inner


@contextmanager
def catch_rdkit_logs():
    log_status = rdBase.LogStatus()
    rdBase.DisableLog("rdApp.*")
    yield
    log_status = {st.split(":")[0]: st.split(":")[1] for st in log_status.split("\n")}
    log_status = {k: True if v == "enabled" else False for k, v in log_status.items()}
    for k, v in log_status.items():
        if v is True:
            rdBase.EnableLog(k)
        else:
            rdBase.DisableLog(k)


@contextmanager
def catch_warning(**kwargs):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", **kwargs)
        yield


def get_centroid(coordinates):
    """Centroid for an array of XYZ coordinates"""
    return np.mean(coordinates, axis=0)


def get_ring_normal_vector(centroid, coordinates):
    """Returns a vector that is normal to the ring plane"""
    # A & B are two edges of the ring
    a = Point3D(*coordinates[0])
    b = Point3D(*coordinates[1])
    # vectors between centroid and these edges
    ca = centroid.DirectionVector(a)
    cb = centroid.DirectionVector(b)
    # cross product between these two vectors
    normal = ca.CrossProduct(cb)
    # cb.CrossProduct(ca) is the normal vector in the opposite direction
    return normal


def angle_between_limits(angle, min_angle, max_angle, ring=False):
    """Checks if an angle value is between min and max angles in radian

    Parameters
    ----------
    angle : float
        Angle to check, in radians
    min_angle : float
        Lower bound angle, in radians
    max_angle : float
        Upper bound angle, in radians
    ring : bool
        Wether the angle being checked involves a ring or not

    Notes
    -----
    When ``ring=True``, the angle is capped between 0 and 90, and so should be
    the min and max angles. This is useful for angles involving a ring's plane
    normal vector.
    """
    if ring:
        if angle >= pi:
            angle %= _90_deg_to_rad
        elif angle > _90_deg_to_rad:
            angle = _90_deg_to_rad - (angle % _90_deg_to_rad)
    return min_angle <= angle <= max_angle


def get_residues_near_ligand(lig, prot, cutoff=6.0):
    """Detects residues close to a reference ligand

    Parameters
    ----------
    lig : prolif.molecule.Molecule
        Select residues that are near this ligand
    prot : prolif.molecule.Molecule
        Protein containing the residues
    cutoff : float
        If any interatomic distance between the ligand reference points and a
        residue is below or equal to this cutoff, the residue will be selected

    Returns
    -------
    residues : list
        A list of unique :class:`~prolif.residue.ResidueId` that are close to
        the ligand
    """
    tree = cKDTree(prot.xyz)
    ix = tree.query_ball_point(lig.xyz, cutoff)
    ix = set([i for lst in ix for i in lst])
    resids = [ResidueId.from_atom(prot.GetAtomWithIdx(i)) for i in ix]
    return list(set(resids))


def split_mol_by_residues(mol):
    """Splits a molecule in multiple fragments based on residues

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        The molecule to fragment

    Returns
    -------
    residues : list
        A list of :class:`rdkit.Chem.rdchem.Mol`

    Notes
    -----
    Code adapted from Maciek Wójcikowski on the RDKit discussion list
    """
    residues = []
    for res in SplitMolByPDBResidues(mol).values():
        for frag in GetMolFrags(res, asMols=True, sanitizeFrags=False):
            # count number of unique residues in the fragment
            resids = {a.GetIdx(): ResidueId.from_atom(a) for a in frag.GetAtoms()}
            if len(set(resids.values())) > 1:
                # split on peptide bonds
                bonds = [
                    b.GetIdx() for b in frag.GetBonds() if is_peptide_bond(b, resids)
                ]
                mols = FragmentOnBonds(frag, bonds, addDummies=False)
                mols = GetMolFrags(mols, asMols=True, sanitizeFrags=False)
                residues.extend(mols)
            else:
                residues.append(frag)
    return residues


def is_peptide_bond(bond, resids):
    """Checks if a bond is a peptide bond based on the ResidueId of the atoms
    on each part of the bond. Also works for disulfide bridges or any bond that
    links two residues in biopolymers.

    Parameters
    ----------
    bond : rdkit.Chem.rdchem.Bond
        The bond to check
    resids : dict
        A dictionnary of ResidueId indexed by atom index
    """
    return resids[bond.GetBeginAtomIdx()] != resids[bond.GetEndAtomIdx()]


def to_dataframe(
    ifp,
    interactions,
    index_col="Frame",
    dtype=None,
    drop_empty=True,
    return_atoms=False,
):
    """Converts IFPs to a pandas DataFrame

    Parameters
    ----------
    ifp : list
        A list of dict in the format {key: bitvector}. "key" is a tuple of
        ligand and protein ResidueId. "bitvector" is either a numpy.ndarray
        of bits, or a list of bitarray, ligand atom indices, and protein atom
        indices. Each dictionnary must also contain an entry that will be used
        as an index, typically a frame number.
    interactions : list
        A list of interactions, in the same order as the bitvector.
    index_col : str
        The dictionnary key that will be used as an index in the DataFrame
    dtype : object or None
        Cast the input of each bit in the bitvector to this type. If None, keep
        the data as is. Not compatible with ``return_atoms=True``
    drop_empty : bool
        Drop columns with only empty values
    return_atoms : bool
        For each residue pair and interaction, return indices of atoms
        responsible for the interaction instead of bits

    Returns
    -------
    df : pandas.DataFrame
        A 3-levels DataFrame where each ligand residue, protein residue, and
        interaction type are in separate columns

    Example
    -------
    ::

        >>> df = prolif.to_dataframe(results, fp.interactions.keys(), dtype=int)
        >>> print(df)
        ligand             LIG1.G
        protein             ILE59                  ILE55       TYR93
        interaction   Hydrophobic HBAcceptor Hydrophobic Hydrophobic PiStacking
        Frame
        0                       0          1           0           0          0
        ...

    .. versionchanged:: 0.3.2
        Moved the ``return_atoms`` parameter from the ``run`` methods to the
        dataframe conversion code
    """
    if dtype and return_atoms:
        raise ValueError("`dtype` cannot be used with `return_atoms=True`")
    ifp = deepcopy(ifp)
    n_interactions = len(interactions)
    empty_value = dtype(False) if dtype else False
    # residue pairs
    keys = sorted(set([k for d in ifp for k in d.keys() if k != index_col]))
    # check if each interaction value is a list of atom indices or smthg else
    has_atom_indices = False
    for d in ifp:
        for key, value in d.items():
            if key != index_col:
                has_atom_indices = isinstance(value[0], Iterable)
                break
    if return_atoms and not has_atom_indices:
        raise ValueError(
            "The IFP either doesn't contain atom indices or is formatted incorrectly"
        )
    # create empty array for each residue pair interaction that doesn't exist
    # in a particular frame
    if has_atom_indices and return_atoms:
        empty_arr = [[None, None]] * n_interactions
    else:
        empty_arr = np.array([empty_value] * n_interactions)
    # sparse to dense
    data = defaultdict(list)
    index = []
    for d in ifp:
        index.append(d.pop(index_col))
        for key in keys:
            try:
                arr = d[key]
            except KeyError:
                data[key].append(empty_arr)
            else:
                if has_atom_indices and return_atoms:
                    arr = list(zip(*arr[1:]))
                elif has_atom_indices:
                    arr = arr[0]
                data[key].append(arr)
    index = pd.Series(index, name=index_col)
    # create dataframes
    if not data:
        warnings.warn("No interaction detected")
        return pd.DataFrame([], index=index)
    values = np.array(
        [np.hstack([np.ravel(a[i]) for a in data.values()]) for i in range(len(index))]
    )
    if has_atom_indices and return_atoms:
        columns = pd.MultiIndex.from_tuples(
            [
                (str(k[0]), str(k[1]), i, a)
                for k in keys
                for i in interactions
                for a in ["ligand", "protein"]
            ],
            names=["ligand", "protein", "interaction", "atom"],
        )
    else:
        columns = pd.MultiIndex.from_tuples(
            [(str(k[0]), str(k[1]), i) for k in keys for i in interactions],
            names=["ligand", "protein", "interaction"],
        )
    df = pd.DataFrame(values, columns=columns, index=index)
    if has_atom_indices and return_atoms:
        df = df.groupby(axis=1, level=["ligand", "protein", "interaction"]).agg(tuple)
    if dtype:
        df = df.astype(dtype)
    if drop_empty:
        if has_atom_indices and return_atoms:
            mask = df.apply(lambda s: ~(s.isin([(None, None)]).all()), axis=0)
        else:
            mask = (df != empty_value).any(axis=0)
        df = df.loc[:, mask]
    return df


def pandas_series_to_bv(s):
    bv = ExplicitBitVect(len(s))
    on_bits = np.where(s >= True)[0].tolist()
    bv.SetBitsFromList(on_bits)
    return bv


def to_bitvectors(df):
    """Converts an interaction DataFrame to a list of RDKit ExplicitBitVector

    Parameters
    ----------
    df : pandas.DataFrame
        A DataFrame where each column corresponds to an interaction between two
        residues

    Returns
    -------
    bv : list
        A list of :class:`~rdkit.DataStructs.cDataStructs.ExplicitBitVect`
        for each frame

    Example
    -------
    ::

        >>> from rdkit.DataStructs import TanimotoSimilarity
        >>> bv = prolif.to_bitvectors(df)
        >>> TanimotoSimilarity(bv[0], bv[1])
        0.42

    """
    return df.apply(pandas_series_to_bv, axis=1).tolist()



from . import datafiles
from ._version import get_versions
from .fingerprint import Fingerprint
from .molecule import Molecule, mol2_supplier, pdbqt_supplier, sdf_supplier
from .residue import ResidueId
from .utils import get_residues_near_ligand, to_bitvectors, to_dataframe

__version__ = get_versions()["version"]
del get_versions



import multiprocessing as mp
from ctypes import c_int32
from time import sleep

from tqdm.auto import tqdm

from .molecule import Molecule


def process_chunk(args):
    """Generates a fingerprint for a chunk of frame indices"""
    traj, lig, prot, chunk = args
    lig_kwargs, prot_kwargs = converter_kwargs
    ifp = []
    for ts in traj[chunk]:
        lig_mol = Molecule.from_mda(lig, **lig_kwargs)
        prot_mol = Molecule.from_mda(prot, **prot_kwargs)
        data = fp.generate(lig_mol, prot_mol, residues=residues, return_atoms=True)
        data["Frame"] = ts.frame
        ifp.append(data)
        if display_progress:
            with pcount.lock:
                pcount.counter.value += 1
    return ifp


def declare_shared_objs_for_chunk(
    fingerprint, resid_list, show_progressbar, progress_counter, rdkitconverter_kwargs
):
    """Declares global objects that are available to the pool of workers for
    a trajectory"""
    global fp, residues, display_progress, pcount, converter_kwargs
    fp = fingerprint
    residues = resid_list
    display_progress = show_progressbar
    pcount = progress_counter
    converter_kwargs = rdkitconverter_kwargs


def process_mol(args):
    """Generates a fingerprint for a single molecule"""
    index, mol = args
    data = fp.generate(mol, prot_mol, residues=residues, return_atoms=True)
    data["Frame"] = index
    return data


def declare_shared_objs_for_mol(fingerprint, pmol, resid_list):
    """Declares global objects that are available to the pool of workers for
    an interable of ligands"""
    global fp, residues, prot_mol
    fp = fingerprint
    residues = resid_list
    prot_mol = pmol


class ProgressCounter:
    """Tracks the progress of the fingerprint analysis accross the pool of
    workers"""

    def __init__(self):
        self.lock = mp.Lock()
        self.counter = mp.Value(c_int32)


class Progress:
    """Handles tracking the progress of the ProgressCounter and updating the
    tqdm progress bar, from within an independent thread"""

    def __init__(self, pcount, *args, **kwargs):
        self.pbar = tqdm(*args, **kwargs)
        self.pcount = pcount

    def __call__(self):
        while self.pbar.n < self.pbar.total:
            if self.pcount.counter.value != 0:
                with self.pcount.lock:
                    n_processed = self.pcount.counter.value
                    self.pcount.counter.value = 0
                self.pbar.update(n_processed)
            sleep(1)



"""
Calculate a Protein-Ligand Interaction Fingerprint --- :mod:`prolif.fingerprint`
================================================================================

.. ipython:: python
    :okwarning:

    import MDAnalysis as mda
    from rdkit.DataStructs import TanimotoSimilarity
    import prolif
    u = mda.Universe(prolif.datafiles.TOP, prolif.datafiles.TRAJ)
    prot = u.select_atoms("protein")
    lig = u.select_atoms("resname LIG")
    fp = prolif.Fingerprint(["HBDonor", "HBAcceptor", "PiStacking", "CationPi",
                             "Cationic"])
    fp.run(u.trajectory[::10], lig, prot)
    df = fp.to_dataframe()
    df
    bv = fp.to_bitvectors()
    TanimotoSimilarity(bv[0], bv[1])
    # save results
    fp.to_pickle("fingerprint.pkl")
    # load
    fp = prolif.Fingerprint.from_pickle("fingerprint.pkl")
    df.equals(fp.to_dataframe())

"""
import multiprocessing as mp
import pickle
from collections.abc import Iterable
from inspect import isgenerator
from threading import Thread

import numpy as np
from rdkit import Chem
from tqdm.auto import tqdm

from .interactions import _INTERACTIONS
from .molecule import Molecule
from .parallel import (
    Progress,
    ProgressCounter,
    declare_shared_objs_for_chunk,
    declare_shared_objs_for_mol,
    process_chunk,
    process_mol,
)
from .utils import get_residues_near_ligand, to_bitvectors, to_dataframe


class _InteractionWrapper:
    """Modifies the return signature of an interaction ``detect`` method by
    forcing it to return only the first element when multiple values are
    returned.

    Raises
    ------
    TypeError
        If the ``detect`` method doesn't return three values

    Notes
    -----
    The original return signature of the decorated function is still accessible
    by calling ``function.__wrapped__(*args, **kwargs)``.

    Example
    -------
    ::

        >>> class Foo:
        ...     def detect(self, *args):
        ...         return 1, 2, 3
        ...
        >>> foo = Foo().detect
        >>> bar = _InteractionWrapper(foo)
        >>> foo()
        (1, 2, 3)
        >>> bar()
        1
        >>> bar.__wrapped__()
        (1, 2, 3)

    .. versionchanged:: 0.3.3
        The function now must return three values

    .. versionchanged:: 1.0.0
        Changed from a wrapper function to a class for easier pickling support

    """

    def __init__(self, func):
        self.__wrapped__ = func
        # add docstring to descriptor
        self.__doc__ = func

    def __repr__(self):  # pragma: no cover
        cls = self.__wrapped__.__self__.__class__
        return f"<{cls.__module__}.{cls.__name__} at {id(self):#x}>"

    def __call__(self, *args, **kwargs):
        results = self.__wrapped__(*args, **kwargs)
        try:
            bool_, lig_idx, prot_idx = results
        except (TypeError, ValueError):
            raise TypeError(
                "Incorrect function signature: the interaction class must "
                "return 3 values (boolean, int, int)"
            ) from None
        return bool_


class Fingerprint:
    """Class that generates an interaction fingerprint between two molecules

    While in most cases the fingerprint will be generated between a ligand and
    a protein, it is also possible to use this class for protein-protein
    interactions, or host-guest systems.

    Parameters
    ----------
    interactions : list
        List of names (str) of interaction classes as found in the
        :mod:`prolif.interactions` module

    Attributes
    ----------
    interactions : dict
        Dictionnary of interaction functions indexed by class name. For more
        details, see :mod:`prolif.interactions`
    n_interactions : int
        Number of interaction functions registered by the fingerprint
    ifp : list, optionnal
        List of interactions fingerprints for the given trajectory.

    Raises
    ------
    NameError : Unknown interaction in the ``interactions`` parameter

    Notes
    -----
    You can use the fingerprint generator in multiple ways:

    - On a trajectory directly from MDAnalysis objects:

    .. ipython:: python

        prot = u.select_atoms("protein")
        lig = u.select_atoms("resname LIG")
        fp = prolif.Fingerprint(["HBDonor", "HBAcceptor", "PiStacking",
                                 "Hydrophobic"])
        fp.run(u.trajectory[:5], lig, prot)
        fp.to_dataframe()

    - On two single structures (from RDKit or MDAnalysis):

    .. ipython:: python

        u.trajectory[0]  # use coordinates of the first frame
        prot = prolif.Molecule.from_mda(prot)
        lig = prolif.Molecule.from_mda(lig)
        ifp = fp.generate(lig, prot)
        ifp["Frame"] = 0
        prolif.to_dataframe([ifp], fp.interactions.keys())

    - On a specific pair of residues for a specific interaction:

    .. ipython:: python

        # ligand-protein
        fp.hbdonor(lig, prot["ASP129.A"])
        # protein-protein (alpha helix)
        fp.hbacceptor(prot["ASP129.A"], prot["CYS133.A"])

    You can also obtain the indices of atoms responsible for the interaction:

    .. ipython:: python

        fp.bitvector_atoms(lig, prot["ASP129.A"])
        fp.hbdonor.__wrapped__(lig, prot["ASP129.A"])


    .. versionchanged:: 1.0.0
        Added pickle support

    """

    def __init__(
        self,
        interactions=[
            "Hydrophobic",
            "HBDonor",
            "HBAcceptor",
            "PiStacking",
            "Anionic",
            "Cationic",
            "CationPi",
            "PiCation",
            "VdWContact",
        ],
    ):
        self._set_interactions(interactions)

    def _set_interactions(self, interactions):
        # read interactions to compute
        self.interactions = {}
        if interactions == "all":
            interactions = self.list_available()
        # sanity check
        unsafe = set(interactions)
        unk = unsafe.symmetric_difference(_INTERACTIONS.keys()) & unsafe
        if unk:
            raise NameError(f"Unknown interaction(s): {', '.join(unk)}")
        # add interaction methods
        for name, interaction_cls in _INTERACTIONS.items():
            if name.startswith("_") or name == "Interaction":
                continue
            func = interaction_cls().detect
            func = _InteractionWrapper(func)
            setattr(self, name.lower(), func)
            if name in interactions:
                self.interactions[name] = func

    def __repr__(self):  # pragma: no cover
        name = ".".join([self.__class__.__module__, self.__class__.__name__])
        params = f"{self.n_interactions} interactions: {list(self.interactions.keys())}"
        return f"<{name}: {params} at {id(self):#x}>"

    @staticmethod
    def list_available(show_hidden=False):
        """List interactions available to the Fingerprint class.

        Parameters
        ----------
        show_hidden : bool
            Show hidden classes (usually base classes whose name starts with an
            underscore ``_``. Those are not supposed to be called directly)
        """
        if show_hidden:
            interactions = [name for name in _INTERACTIONS.keys()]
        else:
            interactions = [
                name
                for name in _INTERACTIONS.keys()
                if not (name.startswith("_") or name == "Interaction")
            ]
        return sorted(interactions)

    @property
    def n_interactions(self):
        return len(self.interactions)

    def bitvector(self, res1, res2):
        """Generates the complete bitvector for the interactions between two
        residues. To get the indices of atoms responsible for each interaction,
        see :meth:`bitvector_atoms`

        Parameters
        ----------
        res1 : prolif.residue.Residue or prolif.molecule.Molecule
            A residue, usually from a ligand
        res2 : prolif.residue.Residue or prolif.molecule.Molecule
            A residue, usually from a protein

        Returns
        -------
        bitvector : numpy.ndarray
            An array storing the encoded interactions between res1 and res2
        """
        bitvector = []
        for func in self.interactions.values():
            bit = func(res1, res2)
            bitvector.append(bit)
        return np.array(bitvector)

    def bitvector_atoms(self, res1, res2):
        """Generates the complete bitvector for the interactions between two
        residues, and returns the indices of atoms responsible for these
        interactions

        Parameters
        ----------
        res1 : prolif.residue.Residue or prolif.molecule.Molecule
            A residue, usually from a ligand
        res2 : prolif.residue.Residue or prolif.molecule.Molecule
            A residue, usually from a protein

        Returns
        -------
        bitvector : numpy.ndarray
            An array storing the encoded interactions between res1 and res2
        lig_atoms : list
            A list containing indices for the ligand atoms responsible for each
            interaction
        pro_atoms : list
            A list containing indices for the protein atoms responsible for
            each interaction


        .. versionchanged:: 0.3.2
            Atom indices are returned as two separate lists instead of a single
            list of tuples

        """
        bitvector = []
        lig_atoms = []
        prot_atoms = []
        for func in self.interactions.values():
            bit, la, pa = func.__wrapped__(res1, res2)
            bitvector.append(bit)
            lig_atoms.append(la)
            prot_atoms.append(pa)
        bitvector = np.array(bitvector)
        return bitvector, lig_atoms, prot_atoms

    def generate(self, lig, prot, residues=None, return_atoms=False):
        """Generates the interaction fingerprint between 2 molecules

        Parameters
        ----------
        lig : prolif.molecule.Molecule
            Molecule for the ligand
        prot : prolif.molecule.Molecule
            Molecule for the protein
        residues : list or "all" or None
            A list of protein residues (:class:`str`, :class:`int` or
            :class:`~prolif.residue.ResidueId`) to take into account for
            the fingerprint extraction. If ``"all"``, all residues will be
            used. If ``None``, at each frame the
            :func:`~prolif.utils.get_residues_near_ligand` function is used to
            automatically use protein residues that are distant of 6.0 Å or
            less from each ligand residue.
        return_atoms : bool
            For each residue pair and interaction, return indices of atoms
            responsible for the interaction instead of bits.

        Returns
        -------
        ifp : dict
            A dictionnary indexed by ``(ligand, protein)`` residue pairs. The
            format for values will depend on ``return_atoms``:

            - A single bitvector if ``return_atoms=False``
            - A tuple of bitvector, ligand atom indices and protein atom
              indices if ``return_atoms=True``

        Example
        -------
        ::

            >>> u = mda.Universe("complex.pdb")
            >>> lig = prolif.Molecule.from_mda(u, "resname LIG")
            >>> prot = prolif.Molecule.from_mda(u, "protein")
            >>> fp = prolif.Fingerprint()
            >>> ifp = fp.generate(lig, prot)

        .. versionadded:: 0.3.2
        """
        ifp = {}
        resids = residues
        if residues == "all":
            resids = prot.residues.keys()
        for lresid, lres in lig.residues.items():
            if residues is None:
                resids = get_residues_near_ligand(lres, prot)
            for prot_key in resids:
                pres = prot[prot_key]
                key = (lresid, pres.resid)
                if return_atoms:
                    ifp[key] = self.bitvector_atoms(lres, pres)
                else:
                    ifp[key] = self.bitvector(lres, pres)
        return ifp

    def run(
        self,
        traj,
        lig,
        prot,
        residues=None,
        converter_kwargs=None,
        progress=True,
        n_jobs=None,
    ):
        """Generates the fingerprint on a trajectory for a ligand and a protein

        Parameters
        ----------
        traj : MDAnalysis.coordinates.base.ProtoReader or MDAnalysis.coordinates.base.FrameIteratorSliced
            Iterate over this Universe trajectory or sliced trajectory object
            to extract the frames used for the fingerprint extraction
        lig : MDAnalysis.core.groups.AtomGroup
            An MDAnalysis AtomGroup for the ligand
        prot : MDAnalysis.core.groups.AtomGroup
            An MDAnalysis AtomGroup for the protein (with multiple residues)
        residues : list or "all" or None
            A list of protein residues (:class:`str`, :class:`int` or
            :class:`~prolif.residue.ResidueId`) to take into account for
            the fingerprint extraction. If ``"all"``, all residues will be
            used. If ``None``, at each frame the
            :func:`~prolif.utils.get_residues_near_ligand` function is used to
            automatically use protein residues that are distant of 6.0 Å or
            less from each ligand residue.
        converter_kwargs : list or None
            List of kwargs passed to the underlying :class:`~MDAnalysis.converters.RDKit.RDKitConverter`
            from MDAnalysis: the first for the ligand, and the second for the protein
        progress : bool
            Use the `tqdm <https://tqdm.github.io/>`_ package to display a
            progressbar while running the calculation
        n_jobs : int or None
            Number of processes to run in parallel. If ``n_jobs=None``, the
            analysis will use all available CPU threads, while if ``n_jobs=1``,
            the analysis will run in serial.

        Raises
        ------
        ValueError : if ``n_jobs <= 0``

        Returns
        -------
        prolif.fingerprint.Fingerprint
            The Fingerprint instance that generated the fingerprint

        Example
        -------
        ::

            >>> u = mda.Universe("top.pdb", "traj.nc")
            >>> lig = u.select_atoms("resname LIG")
            >>> prot = u.select_atoms("protein")
            >>> fp = prolif.Fingerprint().run(u.trajectory[:10], lig, prot)

        .. seealso::

            - :meth:`Fingerprint.generate` to generate the fingerprint between
              two single structures.
            - :meth:`Fingerprint.run_from_iterable` to generate the fingerprint
              between a protein and a collection of ligands.

        .. versionchanged:: 0.3.2
            Moved the ``return_atoms`` parameter from the ``run`` method to the
            dataframe conversion code

        .. versionchanged:: 1.0.0
            Added support for multiprocessing

        .. versionchanged:: 1.1.0
            Added support for passing kwargs to the RDKitConverter through
            the ``converter_kwargs`` parameter

        """
        if n_jobs is not None and n_jobs < 1:
            raise ValueError("n_jobs must be > 0 or None")
        if converter_kwargs is not None and len(converter_kwargs) != 2:
            raise ValueError("converter_kwargs must be a list of 2 dicts")
        if n_jobs != 1:
            return self._run_parallel(
                traj,
                lig,
                prot,
                residues=residues,
                converter_kwargs=converter_kwargs,
                progress=progress,
                n_jobs=n_jobs,
            )
        lig_kwargs, prot_kwargs = converter_kwargs or ({}, {})

        iterator = tqdm(traj) if progress else traj
        if residues == "all":
            residues = Molecule.from_mda(prot, **prot_kwargs).residues.keys()
        ifp = []
        for ts in iterator:
            prot_mol = Molecule.from_mda(prot, **prot_kwargs)
            lig_mol = Molecule.from_mda(lig, **lig_kwargs)
            data = self.generate(
                lig_mol, prot_mol, residues=residues, return_atoms=True
            )
            data["Frame"] = ts.frame
            ifp.append(data)
        self.ifp = ifp
        return self

    def _run_parallel(
        self,
        traj,
        lig,
        prot,
        residues=None,
        converter_kwargs=None,
        progress=True,
        n_jobs=None,
    ):
        """Parallel implementation of :meth:`~Fingerprint.run`"""
        n_chunks = n_jobs if n_jobs else mp.cpu_count()
        try:
            n_frames = traj.n_frames
        except AttributeError:
            # sliced trajectory
            frames = range(traj.start, traj.stop, traj.step)
            traj = lig.universe.trajectory
        else:
            frames = range(n_frames)
        lig_kwargs, prot_kwargs = converter_kwargs or ({}, {})

        if residues == "all":
            residues = Molecule.from_mda(prot, **prot_kwargs).residues.keys()
        chunks = np.array_split(frames, n_chunks)
        # setup parallel progress bar
        pcount = ProgressCounter()
        if progress:
            pbar = Progress(pcount, total=len(frames))
        else:
            pbar = lambda: None
        pbar_thread = Thread(target=pbar, daemon=True)

        # run pool of workers
        with mp.Pool(
            n_jobs,
            initializer=declare_shared_objs_for_chunk,
            initargs=(self, residues, progress, pcount, (lig_kwargs, prot_kwargs)),
        ) as pool:
            pbar_thread.start()
            args = ((traj, lig, prot, chunk) for chunk in chunks)
            results = []
            for ifp in pool.imap_unordered(process_chunk, args):
                results.extend(ifp)
        results.sort(key=lambda ifp: ifp["Frame"])
        self.ifp = results
        return self

    def run_from_iterable(
        self, lig_iterable, prot_mol, residues=None, progress=True, n_jobs=None
    ):
        """Generates the fingerprint between a list of ligands and a protein

        Parameters
        ----------
        lig_iterable : list or generator
            An iterable yielding ligands as :class:`~prolif.molecule.Molecule`
            objects
        prot_mol : prolif.molecule.Molecule
            The protein
        residues : list or "all" or None
            A list of protein residues (:class:`str`, :class:`int` or
            :class:`~prolif.residue.ResidueId`) to take into account for
            the fingerprint extraction. If ``"all"``, all residues will be
            used. If ``None``, at each frame the
            :func:`~prolif.utils.get_residues_near_ligand` function is used to
            automatically use protein residues that are distant of 6.0 Å or
            less from each ligand residue.
        progress : bool
            Use the `tqdm <https://tqdm.github.io/>`_ package to display a
            progressbar while running the calculation
        n_jobs : int or None
            Number of processes to run in parallel. If ``n_jobs=None``, the
            analysis will use all available CPU threads, while if ``n_jobs=1``,
            the analysis will run in serial.

        Raises
        ------
        ValueError : if ``n_jobs <= 0``

        Returns
        -------
        prolif.fingerprint.Fingerprint
            The Fingerprint instance that generated the fingerprint

        Example
        -------
        ::

            >>> prot = mda.Universe("protein.pdb")
            >>> prot = prolif.Molecule.from_mda(prot)
            >>> lig_iter = prolif.mol2_supplier("docking_output.mol2")
            >>> fp = prolif.Fingerprint()
            >>> fp.run_from_iterable(lig_iter, prot)

        .. seealso::

            :meth:`Fingerprint.generate` to generate the fingerprint between
            two single structures

        .. versionchanged:: 0.3.2
            Moved the ``return_atoms`` parameter from the ``run_from_iterable``
            method to the dataframe conversion code

        .. versionchanged:: 1.0.0
            Added support for multiprocessing

        """
        if n_jobs is not None and n_jobs < 1:
            raise ValueError("n_jobs must be > 0 or None")
        if n_jobs != 1:
            return self._run_iter_parallel(
                lig_iterable=lig_iterable,
                prot_mol=prot_mol,
                residues=residues,
                progress=progress,
                n_jobs=n_jobs,
            )

        iterator = tqdm(lig_iterable) if progress else lig_iterable
        if residues == "all":
            residues = prot_mol.residues.keys()
        ifp = []
        for i, lig_mol in enumerate(iterator):
            data = self.generate(
                lig_mol, prot_mol, residues=residues, return_atoms=True
            )
            data["Frame"] = i
            ifp.append(data)
        self.ifp = ifp
        return self

    def _run_iter_parallel(
        self, lig_iterable, prot_mol, residues=None, progress=True, n_jobs=None
    ):
        """Parallel implementation of :meth:`~Fingerprint.run_from_iterable`"""
        if isinstance(lig_iterable, Chem.SDMolSupplier) or (
            isinstance(lig_iterable, Iterable) and not isgenerator(lig_iterable)
        ):
            total = len(lig_iterable)
        else:
            total = None
        suppl = enumerate(lig_iterable)
        if residues == "all":
            residues = prot_mol.residues.keys()

        with mp.Pool(
            n_jobs,
            initializer=declare_shared_objs_for_mol,
            initargs=(self, prot_mol, residues),
        ) as pool:
            results = []
            for data in tqdm(
                pool.imap_unordered(process_mol, suppl),
                total=total,
                disable=not progress,
            ):
                results.append(data)
        results.sort(key=lambda ifp: ifp["Frame"])
        self.ifp = results
        return self

    def to_dataframe(self, **kwargs):
        """Converts fingerprints to a pandas DataFrame

        Parameters
        ----------
        dtype : object or None
            Cast the input of each bit in the bitvector to this type. If None,
            keep the data as is
        drop_empty : bool
            Drop columns with only empty values
        return_atoms : bool
            For each residue pair and interaction, return indices of atoms
            responsible for the interaction instead of bits

        Returns
        -------
        df : pandas.DataFrame
            A DataFrame storing the results frame by frame and residue by
            residue. See :meth:`~prolif.utils.to_dataframe` for more
            information

        Raises
        ------
        AttributeError
            If the :meth:`run` method hasn't been used

        Example
        -------
        ::

            >>> df = fp.to_dataframe(dtype=np.uint8)
            >>> print(df)
            ligand             LIG1.G
            protein             ILE59                  ILE55       TYR93
            interaction   Hydrophobic HBAcceptor Hydrophobic Hydrophobic PiStacking
            Frame
            0                       0          1           0           0          0
            ...

        """
        if hasattr(self, "ifp"):
            return to_dataframe(self.ifp, self.interactions.keys(), **kwargs)
        raise AttributeError("Please use the `run` method before")

    def to_bitvectors(self):
        """Converts fingerprints to a list of RDKit ExplicitBitVector

        Returns
        -------
        bvs : list
            A list of :class:`~rdkit.DataStructs.cDataStructs.ExplicitBitVect`
            for each frame

        Raises
        ------
        AttributeError
            If the :meth:`run` method hasn't been used

        Example
        -------
        ::

            >>> from rdkit.DataStructs import TanimotoSimilarity
            >>> bv = fp.to_bitvectors()
            >>> TanimotoSimilarity(bv[0], bv[1])
            0.42

        """
        if hasattr(self, "ifp"):
            df = self.to_dataframe()
            return to_bitvectors(df)
        raise AttributeError("Please use the `run` method before")

    def to_pickle(self, path=None):
        """Dumps the fingerprint object as a pickle.

        Parameters
        ----------
        path : str, pathlib.Path or None
            Output path. If ``None``, the method returns the pickle as bytes.

        Returns
        -------
        obj : None or bytes
            ``None`` if ``path`` is set, else the bytes corresponding to the
            pickle

        Example
        -------
        ::

            >>> dump = fp.to_pickle()
            >>> saved_fp = Fingerprint.from_pickle(dump)
            >>> fp.to_pickle("data/fp.pkl")
            >>> saved_fp = Fingerprint.from_pickle("data/fp.pkl")

        .. seealso::

            :meth:`~Fingerprint.from_pickle` for loading the pickle dump.

        .. versionadded:: 1.0.0

        """
        if path:
            with open(path, "wb") as f:
                pickle.dump(self, f)
        else:
            return pickle.dumps(self)

    @classmethod
    def from_pickle(cls, path_or_bytes):
        """Creates a fingerprint object from a pickle dump.

        Parameters
        ----------
        path_or_bytes : str, pathlib.Path or bytes
            The path to the pickle file, or bytes corresponding to a pickle
            dump

        .. seealso::

            :meth:`~Fingerprint.to_pickle` for creating the pickle dump.

        .. versionadded:: 1.0.0

        """
        if isinstance(path_or_bytes, bytes):
            return pickle.loads(path_or_bytes)
        with open(path_or_bytes, "rb") as f:
            return pickle.load(f)



import os
import re

from setuptools import setup

import versioneer

GITHUB_ACTIONS = os.environ.get("GITHUB_ACTIONS", False)
READTHEDOCS = os.environ.get("READTHEDOCS", False)

# manually check RDKit version
try:
    from rdkit import __version__ as rdkit_version
except ImportError:
    if not (GITHUB_ACTIONS or READTHEDOCS):
        raise ImportError("ProLIF requires RDKit but it is not installed")
else:
    if re.match(r"^(20[0-1][0-9])|(2020)", rdkit_version):
        raise ValueError("ProLIF requires a version of RDKit >= 2021")

setup(version=versioneer.get_version())
